package net.codinux.banking.client

import net.codinux.banking.client.model.*
import net.codinux.banking.client.model.options.RetrieveTransactions
import net.codinux.banking.client.model.request.GetAccountDataRequest
import net.codinux.banking.client.model.request.TransferMoneyRequestForUser
import net.codinux.banking.client.model.response.GetAccountDataResponse
import net.codinux.banking.client.model.response.GetTransactionsResponse
import net.codinux.banking.client.model.response.Response
import net.codinux.banking.client.model.response.TransferMoneyResponse
import net.codinux.banking.client.model.tan.TanMethodType

interface BankingClient {

    /**
     * Retrieves account data like customer name, her bank accounts, their balance and account transactions.
     *
     * By default the account transactions of the last 90 days are retrieved (which in most cases doesn't require a
     * TAN according to PSD2).
     *
     * If you like to retrieve the transactions of a different period, use the method overload that takes a
     * [GetAccountDataRequest] parameter and set [RetrieveTransactions] and may to and from date in its options object.
     */
    suspend fun getAccountDataAsync(bankCode: String, loginName: String, password: String) =
        getAccountDataAsync(GetAccountDataRequest(bankCode, loginName, password))

    /**
     * Retrieves account data like customer name, her bank accounts, their balance and account transactions.
     *
     * By default the account transactions of the last 90 days are retrieved (which in most cases doesn't require a
     * TAN according to PSD2).
     *
     * If you like to retrieve the transactions of a different period, set [RetrieveTransactions] and may to and from
     * date in [GetAccountDataRequest.options].
     */
    suspend fun getAccountDataAsync(request: GetAccountDataRequest): Response<GetAccountDataResponse>

    /**
     * Convenience wrapper around [getAccountDataAsync].
     * Updates account's transactions beginning from [BankAccount.lastAccountUpdateTime].
     * This may requires TAN if [BankAccount.lastAccountUpdateTime] is older than 90 days.
     *
     * Optionally specify which [accounts] should be updated. If not specified all accounts will be updated.
     */
    suspend fun updateAccountTransactionsAsync(
        bank: BankAccess, accounts: List<BankAccount>? = null,
        preferredTanMethodsIfSelectedTanMethodIsNotAvailable: List<TanMethodType>? = TanMethodType.TanMethodsPreferredByMostApplications
    ): Response<List<GetTransactionsResponse>>


    suspend fun transferMoneyAsync(bankCode: String, loginName: String, password: String, recipientName: String,
                                   recipientAccountIdentifier: String, amount: Amount, paymentReference: String? = null) =
        transferMoneyAsync(TransferMoneyRequestForUser(bankCode, loginName, password, null, recipientName, recipientAccountIdentifier, null, amount, paymentReference = paymentReference))

    suspend fun transferMoneyAsync(request: TransferMoneyRequestForUser): Response<TransferMoneyResponse>

}