package net.codinux.banking.client

import net.codinux.banking.client.model.Amount
import net.codinux.banking.client.model.BankAccount
import net.codinux.banking.client.model.options.GetAccountDataOptions
import net.codinux.banking.client.model.options.RetrieveTransactions
import net.codinux.banking.client.model.request.TransferMoneyRequest
import net.codinux.banking.client.model.response.GetAccountDataResponse
import net.codinux.banking.client.model.response.GetTransactionsResponse
import net.codinux.banking.client.model.response.Response
import net.codinux.banking.client.model.response.TransferMoneyResponse
import net.codinux.banking.client.model.tan.TanMethodType

interface BankingClientForUser {

    /**
     * Retrieves account data like customer name, her bank accounts, their balance and account transactions.
     *
     * By default the account transactions of the last 90 days are retrieved (which in most cases doesn't require a
     * TAN according to PSD2).
     *
     * If you like to retrieve the transactions of a different period, use the method overload that takes a
     * [GetAccountDataOptions] parameter and set [RetrieveTransactions] and may to and from date.
     */
    // for languages not supporting default parameters (Java, Swift, JS, ...)
    suspend fun getAccountDataAsync() = getAccountDataAsync(GetAccountDataOptions())

    /**
     * Retrieves account data like customer name, her bank accounts, their balance and account transactions.
     *
     * By default the account transactions of the last 90 days are retrieved (which in most cases doesn't require a
     * TAN according to PSD2).
     *
     * If you like to retrieve the transactions of a different period, set [GetAccountDataOptions.retrieveTransactions]
     * and may [GetAccountDataOptions.retrieveTransactionsFrom] and [GetAccountDataOptions.retrieveTransactionsTo].
     */
    suspend fun getAccountDataAsync(options: GetAccountDataOptions): Response<GetAccountDataResponse>

    /**
     * Convenience wrapper around [getAccountDataAsync].
     * Updates account's transactions beginning from [BankAccount.lastAccountUpdateTime].
     * This may requires TAN if [BankAccount.lastAccountUpdateTime] is older than 90 days.
     */
    suspend fun updateAccountTransactionsAsync(
        accounts: List<BankAccount>? = null,
        preferredTanMethodsIfSelectedTanMethodIsNotAvailable: List<TanMethodType>? = TanMethodType.TanMethodsPreferredByMostApplications
    ): Response<List<GetTransactionsResponse>>


    suspend fun transferMoneyAsync(recipientName: String, recipientAccountIdentifier: String, amount: Amount, paymentReference: String? = null): Response<TransferMoneyResponse>

    suspend fun transferMoneyAsync(request: TransferMoneyRequest): Response<TransferMoneyResponse>

}