package net.codinux.banking.client.model

import kotlinx.datetime.LocalDate
import net.codinux.banking.client.model.config.JsonIgnore
import net.codinux.banking.client.model.config.NoArgConstructor

@Suppress("RUNTIME_ANNOTATION_NOT_SUPPORTED")
@NoArgConstructor
open class AccountTransaction(
    val amount: Amount = Amount.Zero, // TODO: a string is really bad in UI, find a better solution
    val currency: String,
    val reference: String?, // Alternative: Remittance information, Transaction description, (payment) purpose, payment reference

    /**
     * Transaction date (Buchungstag) - der Tag, an dem ein Zahlungsvorgang in das System einer Bank eingegangen ist.
     * Das bedeutet aber nicht automatisch, dass das Geld schon verfügbar ist. Dafür ist die Wertstellung entscheidend.
     */
    val bookingDate: LocalDate,
    /**
     * Effective date (Wertstellung / Valutadatum) - der Tag an dem das Geld verfügbar ist. An diesem Tag wird die
     * Kontobewegung wirksam.
     *
     * Buchung und Wertstellung erfolgen häufig am gleichen Tag, das muss aber nicht immer der Fall sein.
     */
    val valueDate: LocalDate,

    // deutsche Begriffe: "Transaktionspartei" oder "Beteiligte Partei"
    // Englische: Transaction party (ist die beste Wahl für eine neutrale und übergreifende Beschreibung),
    //            Counterparty (ist nützlich in formellen oder finanziellen Kontexten), Participant (ist breiter gefasst, aber weniger präzise)
    /**
     * Name of the Payer (debtor) or Payee (creditor)
     */
    val otherPartyName: String? = null, // Alternatives: Parties involved, Transaction parties.single names: Beneficiary, Payee respectively Payer, Debtor
    /**
     * Bank Identifier, in most cases BIC, of the Payer or Payee
     */
    val otherPartyBankId: String? = null,
    /**
     * Account Identifier, in most cases IBAN, of the Payer or Payee
     */
    val otherPartyAccountId: String? = null,

    /**
     * Buchungstext, z. B. DAUERAUFTRAG, BARGELDAUSZAHLUNG, ONLINE-UEBERWEISUNG, FOLGELASTSCHRIFT, ...
     */
    val postingText: String? = null,

    val openingBalance: Amount? = null,
    val closingBalance: Amount? = null,

    /**
     * Auszugsnummer
     */
    val statementNumber: Int? = null,
    /**
     * Blattnummer
     */
    val sheetNumber: Int? = null,

    /**
     * Kundenreferenz.
     */
    val customerReference: String? = null,
    /**
     * Bankreferenz
     */
    val bankReference: String? = null,
    /**
     * Währungsart und Umsatzbetrag in Ursprungswährung
     */
    val furtherInformation: String? = null,


    /*      Information mostly of direct debit (Lastschrift)       */

    val endToEndReference: String? = null,
    val mandateReference: String? = null,
    val creditorIdentifier: String? = null,
    val originatorsIdentificationCode: String? = null,

    /**
     * Summe aus Auslagenersatz und Bearbeitungsprovision bei einer nationalen Rücklastschrift
     * sowie optionalem Zinsausgleich.
     */
    val compensationAmount: String? = null,
    /**
     * Betrag der ursprünglichen Lastschrift
     */
    val originalAmount: String? = null,
    /**
     * Abweichender Überweisender oder Zahlungsempfänger
     */
    val deviantOriginator: String? = null,
    /**
     * Abweichender Zahlungsempfänger oder Zahlungspflichtiger
     */
    val deviantRecipient: String? = null,
    val referenceWithNoSpecialType: String? = null,

    /**
     * Primanoten-Nr.
     */
    val journalNumber: String? = null,
    /**
     * Bei R-Transaktionen siehe Tabelle der
     * SEPA-Rückgabecodes, bei SEPALastschriften siehe optionale Belegung
     * bei GVC 104 und GVC 105 (GVC = Geschäftsvorfallcode)
     */
    val textKeyAddition: String? = null,

    /**
     * Referenznummer, die vom Sender als eindeutige Kennung für die Nachricht vergeben wurde
     * (z.B. als Referenz auf stornierte Nachrichten).
     */
    val orderReferenceNumber: String? = null,
    /**
     * Bezugsreferenz
     */
    val referenceNumber: String? = null,

    /**
     * Storno, ob die Buchung storniert wurde(?).
     * Aus:
     * „RC“ = Storno Haben
     * „RD“ = Storno Soll
     */
    val isReversal: Boolean = false,

    var userSetReference: String? = null,
    var userSetOtherPartyName: String? = null,
    var category: String? = null,
    var notes: String? = null,
) {

    open val identifier by lazy {
        "$amount $currency $bookingDate $valueDate $reference $otherPartyName $otherPartyBankId $otherPartyAccountId"
    }

    @get:JsonIgnore
    open val displayedReference: String?
        get() = userSetReference ?: referenceNumber

    @get:JsonIgnore
    open val displayedOtherPartyName: String?
        get() = userSetOtherPartyName ?: otherPartyName

    @get:JsonIgnore
    open val displayedOtherPartyNameOrPostingText: String?
        get() = displayedOtherPartyName ?: postingText


    override fun toString() = "${valueDate.dayOfMonth}.${valueDate.monthNumber}.${valueDate.year} ${amount.toString().padStart(4, ' ')} ${if (currency == "EUR") "€" else currency} ${otherPartyName ?: ""} - $reference"
}