package net.codinux.banking.client.model

import net.codinux.banking.client.model.config.JsonIgnore
import net.codinux.banking.client.model.config.NoArgConstructor
import net.codinux.banking.client.model.tan.TanMedium
import net.codinux.banking.client.model.tan.TanMethod

@Suppress("RUNTIME_ANNOTATION_NOT_SUPPORTED")
@NoArgConstructor
open class BankAccess(
    /**
     * The country specific bank code, like in Germany the Bankleitzahl, in USA the Routing Number, in Great Britain
     * the Sort Code, in India the FSC Code, ...
     */
    val domesticBankCode: String,
    var loginName: String,
    /**
     * User may decides to not save password.
     */
    var password: String?,

    val bankName: String,
    val bic: String?, // not all banks (in the world) have a BIC

    val customerName: String,
    /**
     * The customer is the person or organisation that owns the account at the bank.
     *
     * The user is the person that access a customer's account via e.g. FinTS. The user may but not necessarily is
     * identical with the customer. E.g. an employee from the bookkeeping departement or a tax consultant may only
     * access an account for someone else.
     *
     * So in most cases the userId is identical with the customerId = loginName in our speech, but there are rare cases
     * where the userId differs from customerId.
     */
    var userId: String? = null,

    open val accounts: List<BankAccount> = emptyList(),

    /**
     * Identifier of selected TanMethod.
     *
     * As [tanMethods] also contains selected TanMethod, we didn't want to duplicate this object. Use
     * [selectedTanMethod] to get selected TanMethod or iterate over [tanMethods] and filter selected one by this id.
     */
    var selectedTanMethodIdentifier: String? = null,
    open val tanMethods: MutableList<out TanMethod> = mutableListOf(),

    /**
     * Identifier of selected TanMedium.
     *
     * As [tanMedia] also contains selected TanMedium, we didn't want to duplicate this object. Use [selectedTanMedium]
     * to get selected TanMedium or iterate over [tanMedia] and filter selected one by this medium name.
     */
    var selectedTanMediumIdentifier: String? = null,
    open val tanMedia: MutableList<out TanMedium> = mutableListOf(),

    var bankingGroup: BankingGroup? = null,
    open var serverAddress: String? = null,
    /**
     * The ISO code of the country where the bank resides and to know the system of [domesticBankCode].
     */
    val countryCode: String = "de"
) {

    var userSetDisplayName: String? = null
    var displayIndex: Int = 0

    var iconUrl: String? = null

    var wrongCredentialsEntered: Boolean = false

    /**
     * BankingClient specific data of this account that the client needs to fulfill its job.
     *
     * You should treat it as opaque data, that only makes sense to the BankingClient, and pass it back to the client if set.
     *
     * For fints4k e.g. contains the FinTS jobs the bank supports, FinTS specific data like KundensystemID and so on.
     *
     * The deserialized in-memory only value of [serializedClientData] so that we don't have to deserialize [serializedClientData] each time.
     */
    var clientData: Any? = null

    /**
     * Serialized version of [clientData].
     *
     * The same as with [clientData] you should treat this value as opaque that only makes sense to the client implementation.
     *
     * [clientData] is the deserialized in-memory model of this value, so that we don't have to serialize this value each time.
     * serializedClientData is the serialized version of clientData so that you can store (and restore) it e.g. to your
     * database, so that on next application start client implementation doesn't have to fetch all these data again.
     * Speeds up e.g. getting account transactions and transferring money with fints4k as FinTS requires quite a lot of
     * data before account transactions can be retrieved.
     */
    var serializedClientData: String? = null


    @get:JsonIgnore
    open val displayName: String
        get() = userSetDisplayName ?: bankName


    @get:JsonIgnore
    open val accountsSorted: List<out BankAccount>
        get() = accounts.sortedBy { it.displayIndex }

    @get:JsonIgnore
    open val tanMethodsSorted: List<out TanMethod>
        get() = tanMethods.sortedBy { it.identifier }

    @get:JsonIgnore
    open val tanMediaSorted: List<out TanMedium>
        get() = tanMedia.sortedBy { it.status }


    @get:JsonIgnore
    val selectedTanMethod: TanMethod
        get() = tanMethods.first { it.identifier == selectedTanMethodIdentifier }

    @get:JsonIgnore
    val selectedTanMedium: TanMedium?
        get() = tanMedia.firstOrNull { it.identifier == selectedTanMediumIdentifier }


    override fun toString() = "$bankName $loginName, ${accounts.size} accounts"

}