package net.codinux.banking.client.model

import kotlinx.datetime.*
import net.codinux.banking.client.model.config.JsonIgnore
import net.codinux.banking.client.model.config.NoArgConstructor
import net.codinux.banking.client.model.securitiesaccount.Holding

@Suppress("RUNTIME_ANNOTATION_NOT_SUPPORTED")
@NoArgConstructor
open class BankAccount(
    val identifier: String,
    val subAccountNumber: String? = null,
    val iban: String? = null,
    val productName: String? = null,
    var accountHolderName: String,
    val type: BankAccountType = BankAccountType.Other,
    val currency: String = DefaultValues.DefaultCurrency,
    var accountLimit: String? = null,

    val isAccountTypeSupportedByApplication: Boolean = false,
    val features: Set<BankAccountFeatures> = emptySet(),

    var balance: Amount = Amount.Zero,

    val serverTransactionsRetentionDays: Int? = null,
    open var lastAccountUpdateTime: Instant? = null,
    var retrievedTransactionsFrom: LocalDate? = null,

    open val bookedTransactions: MutableList<out AccountTransaction> = mutableListOf(),
    open val prebookedTransactions: MutableList<out PrebookedAccountTransaction> = mutableListOf(),
    open val holdings: MutableList<out Holding> = mutableListOf(),

    var userSetDisplayName: String? = null,
    var displayIndex: Int = 0,

    var hideAccount: Boolean = false,
    var includeInAutomaticAccountsUpdate: Boolean = true
) {
    @get:JsonIgnore
    open val displayName: String
        get() = userSetDisplayName ?: productName ?: identifier

    open fun addTransactions(transactions: List<out AccountTransaction>) {
        (this.bookedTransactions as MutableList<AccountTransaction>).addAll(transactions)
    }

    open fun addPrebookedTransactions(transactions: List<out PrebookedAccountTransaction>) {
        (this.prebookedTransactions as MutableList<PrebookedAccountTransaction>).addAll(transactions)
    }

    open fun addHoldings(holdings: List<out Holding>) {
        (this.holdings as MutableList<Holding>).addAll(holdings)
    }

    @get:JsonIgnore
    open val supportsBalanceRetrieval: Boolean
        get() = supportsAnyFeature(BankAccountFeatures.RetrieveBalance)

    @get:JsonIgnore
    open val supportsTransactionRetrieval: Boolean
        get() = supportsAnyFeature(BankAccountFeatures.RetrieveTransactions)

    @get:JsonIgnore
    open val supportsMoneyTransfer: Boolean
        get() = supportsAnyFeature(BankAccountFeatures.TransferMoney)

    @get:JsonIgnore
    open val supportsInstantTransfer: Boolean
        get() = supportsAnyFeature(BankAccountFeatures.InstantTransfer)

    open fun supportsAnyFeature(vararg features: BankAccountFeatures): Boolean =
        features.any { this.features.contains(it) }

    /**
     * Determines if all transactions that are retained on bank server have been fetched.
     *
     * Does this by comparing [serverTransactionsRetentionDays] to [retrievedTransactionsFrom].
     */
    open val haveAllRetainedTransactionsBeenRetrieved: Boolean by lazy {
        val fromDay = retrievedTransactionsFrom
        if (fromDay == null) {
            false
        } else {
            // if countDaysForWhichTransactionsAreKeptOnBankServer is not set, we cannot know for how long bank server keeps transactions. We then assume for 90 days
            val storageDays = serverTransactionsRetentionDays ?: 90
            fromDay < Clock.System.now().toLocalDateTime(TimeZone.of("Europe/Berlin")).date.minus(storageDays, DateTimeUnit.DAY)
        }
    }

    override fun toString() = "$type $identifier $productName (IBAN: $iban)"
}