package net.codinux.banking.client.model.request

import net.codinux.banking.client.model.Amount
import net.codinux.banking.client.model.BankAccountIdentifier
import net.codinux.banking.client.model.DefaultValues
import net.codinux.banking.client.model.config.NoArgConstructor
import net.codinux.banking.client.model.tan.TanMethodType

@NoArgConstructor
open class TransferMoneyRequest(

    /*      Sender settings     */

    /**
     * The account from which the money should be withdrawn.
     * If not specified client retrieves all bank accounts and checks if there is exactly one that supports money transfer.
     * If no or more than one bank account supports money transfer, the error codes NoAccountSupportsMoneyTransfer or MoreThanOneAccountSupportsMoneyTransfer are returned.
     */
    open val senderAccount: BankAccountIdentifier? = null,


    /*      Recipient settings     */

    open val recipientName: String,
    /**
     * The identifier of recipient's account. In most cases the IBAN.
     */
    open val recipientAccountIdentifier: String,
    /**
     * The identifier of recipient's bank. In most cases the BIC.
     * Can be omitted for German banks as the BIC can be derived from IBAN.
     */
    open val recipientBankIdentifier: String? = null,


    /*      Transfer data    */

    open val amount: Amount,

    open val currency: String = DefaultValues.DefaultCurrency,

    /**
     * The purpose of payment. An optional value that tells the reason for the transfer.
     *
     * May not be longer than 140 characters. Some characters are forbidden (TODO: add reference of forbidden characters).
     */
    open val paymentReference: String? = null, // Alternativ: Purpose of payment

    /**
     * If transfer should be executed as 'real-time transfer', that is the money is in less than 10 seconds
     * transferred to the account of the recipient.
     *
     * May costs extra fees.
     *
     * Not supported by all sender and recipient banks.
     */
    open val instantTransfer: Boolean = false, // Alternativ: Instant payment ("Instant payment" ist ebenfalls weit verbreitet und wird oft im Kontext von digitalen Zahlungen verwendet, bei denen die Zahlung in Echtzeit erfolgt. Es kann jedoch breiter gefasst sein und umfasst nicht nur Banktransfers, sondern auch andere Arten von Sofortzahlungen (z.B. mobile Zahlungen).)

    /**
     * Specifies which [TanMethodType] should be preferred when having to choose between multiple available for user
     * without requesting the user to choose one.
     *
     * By default we don't ask the user which TanMethod she prefers but choose one that could match best. If she really
     * likes to use a different one, she can select another one in EnterTanDialog.
     *
     * By default we prefer non visual TanMethods (like AppTan and SMS) over image based TanMethods (like QR-code and
     * photoTan) and exclude ChipTanUsb, which is not supported by application, and FlickerCode, which is hard to
     * implement and therefore most applications have not implemented.
     *
     * Console apps can only handle non visual TanMethods.
     * But also graphical applications prefer non visual TanMethods as then they only have to display a text field to input
     * TAN, and then image based TanMethods as then they additionally only have to display an image.
     */
    val preferredTanMethods: List<TanMethodType>? = TanMethodType.TanMethodsPreferredByMostApplications,

    val tanMethodsNotSupportedByApplication: List<TanMethodType> = TanMethodType.TanMethodsNotSupportedByMostApplications,

    val clientData: Any? = null,
    var serializedClientData: String? = null
) {
    override fun toString() = "$amount to $recipientName - $paymentReference"
}