package net.codinux.banking.client.model.tan

import kotlinx.datetime.Clock
import kotlinx.datetime.Instant
import net.codinux.banking.client.model.BankAccountViewInfo
import net.codinux.banking.client.model.BankAccess
import net.codinux.banking.client.model.BankViewInfo
import net.codinux.banking.client.model.config.JsonIgnore
import net.codinux.banking.client.model.config.NoArgConstructor

@Suppress("RUNTIME_ANNOTATION_NOT_SUPPORTED")
@NoArgConstructor
open class TanChallenge(
    open val type: TanChallengeType,
    open val forAction: ActionRequiringTan,
    open val messageToShowToUser: String,

    /**
     * Identifier of selected TanMethod.
     *
     * As [availableTanMethods] also contains selected TanMethod, we didn't want to duplicate this object. Use
     * [selectedTanMethod] to get selected TanMethod or iterate over [availableTanMethods] and filter selected one by this id.
     */
    open val selectedTanMethodIdentifier: String,
    /**
     * When adding an account, frontend has no UserAccount object in BankingClientCallback to know which TanMethods are
     * available for user.
     * Also on other calls to bank server, bank server may returned an updated list of available TanMethods, so that
     * [BankAccess] may contains an outdated list of available TanMethods.
     *
     * Therefore i added list with up to date TanMethods here to ensure EnterTanDialog can display user's up to date TanMethods.
     */
    open val availableTanMethods: List<TanMethod>,

    /**
     * Identifier of selected TanMedium.
     *
     * As [availableTanMedia] also contains selected TanMedium, we didn't want to duplicate this object. Use
     * [selectedTanMedium] to get selected TanMedium or iterate over [availableTanMedia] and filter selected one by this medium name.
     */
    open val selectedTanMediumIdentifier: String? = null,
    open val availableTanMedia: List<TanMedium> = emptyList(),

    open val tanImage: TanImage? = null,
    open val flickerCode: FlickerCode? = null,
    open val bank: BankViewInfo,
    open val account: BankAccountViewInfo? = null,
    /**
     * Datum und Uhrzeit, bis zu welchem Zeitpunkt eine TAN auf Basis der gesendeten Challenge gültig ist. Nach Ablauf der Gültigkeitsdauer wird die entsprechende TAN entwertet.
     *
     * In server's time zone, that is Europe/Berlin.
     */
    val tanExpirationTime: Instant? = null,
    val challengeCreationTimestamp: Instant = Clock.System.now()
) {

    @get:JsonIgnore
    open val selectedTanMethod: TanMethod
        get() = availableTanMethods.first { it.identifier == selectedTanMethodIdentifier }

    @get:JsonIgnore
    open val selectedTanMedium: TanMedium?
        get() = availableTanMedia.firstOrNull { it.mediumName == selectedTanMediumIdentifier }


    /**
     * Not implemented for all client, only implementing client for now: FinTs4jBankingClient.
     *
     * If a TAN expires either when [TanChallenge.tanExpirationTime] or a default timeout (15 min) is exceeded,
     * you can add a callback to get notified when TAN expired e.g. to close a EnterTanDialog.
     */
    open fun addTanExpiredCallback(callback: () -> Unit) {

    }

    /**
     * Principally a no-op method, not implemented for all client, only implementing client for now: FinTs4jBankingClient.
     *
     * If a TAN is requested for a decoupled TAN method like [TanMethodType.DecoupledTan] or [TanMethodType.DecoupledPushTan],
     * you can add a callback to get notified when user approved TAN in her app e.g. to close a EnterTanDialog.
     */
    open fun addUserApprovedDecoupledTanCallback(callback: () -> Unit) {

    }


    override fun toString(): String {
        return "$selectedTanMethod $forAction: $messageToShowToUser" + when (type) {
            TanChallengeType.EnterTan -> ""
            TanChallengeType.Image -> ", Image: $tanImage"
            TanChallengeType.FlickerCode -> ", FlickerCode: $flickerCode"
        }
    }

}