package net.codinux.banking.client.fints4k.example

import kotlinx.datetime.Clock
import kotlinx.datetime.LocalDate
import kotlinx.datetime.TimeZone
import kotlinx.datetime.toLocalDateTime
import net.codinux.banking.client.SimpleBankingClientCallback
import net.codinux.banking.client.fints4k.FinTs4kBankingClient
import net.codinux.banking.client.fints4k.FinTs4kBankingClientForUser
import net.codinux.banking.client.getAccountData
import net.codinux.banking.client.model.options.GetAccountDataOptions
import net.codinux.banking.client.model.options.RetrieveTransactions
import net.codinux.banking.client.model.response.GetAccountDataResponse
import net.codinux.banking.client.model.response.Response
import net.codinux.banking.client.model.tan.EnterTanResult
import net.codinux.banking.client.service.BankingModelService
import net.codinux.banking.client.updateAccountTransactions
import net.codinux.banking.client.model.extensions.minusDays

fun main() {
    val showUsage = ShowUsage()

    showUsage.getAccountDataSimpleExample()
//    showUsage.getAccountDataFullExample()
}

class ShowUsage {

    private val bankCode = "" // Bankleitzahl deiner Bank

    private val loginName = "" // Online-Banking Login Name mit dem du dich beim Online-Banking deiner Bank anmeldest

    private val password = "" // Online-Banking Password mit dem du dich beim Online-Banking deiner Bank anmeldest


    fun getAccountDataSimpleExample() {
        val client = FinTs4kBankingClient(SimpleBankingClientCallback())

        val response = client.getAccountData(bankCode, loginName, password) // that's all

        printReceivedData(response) // now print retrieved data (save it to database, display it in UI, ...)
    }

    fun getAccountDataFullExample() {
        println("Enter password for $bankCode:")
        val password = readln() // as an alternative for hard coding password; of course can also be done for bankCode and login name

        val client = FinTs4kBankingClientForUser(bankCode, loginName, password, SimpleBankingClientCallback { tanChallenge, callback ->
            println("A TAN is required for ${tanChallenge.forAction}. Selected TAN method is '${tanChallenge.selectedTanMethod.displayName}'. Messsage of your credit institute:")
            println(tanChallenge.messageToShowToUser)
            println("Get TAN from your TAN app etc., enter it and press Enter (or press Enter without an input to abort process):")

            val tan: String? = readlnOrNull().takeUnless { it.isNullOrBlank() } // map empty input to null to abort process
            callback.invoke(EnterTanResult(tan))
        })

        val options = GetAccountDataOptions(
            retrieveBalance = true, // retrieve balance (Saldo / Kontostand) yes or no
            retrieveTransactions = RetrieveTransactions.AccordingToRetrieveFromAndTo, // so that fromDate and toDate below determine of which time period transactions (Umsätze) should be retrieved; defaults to OfLast90Days which in most cases doesn't require a TAN
            retrieveTransactionsFrom = LocalDate(2023, 1, 1),
            retrieveTransactionsTo = LocalDate(2023, 12, 31),
            abortIfTanIsRequired = false // e.g. for command line application when entering TAN is either not possible or a TAN procedure is used that cannot be handled via a break point (e.g. showing a TAN image or flicker code)
        )
        val response = client.getAccountData(options)

        response.error?.let{ error ->
            println("Could not fetch account data: ${error.type} ${error.internalError ?: error.errorMessagesFromBank.joinToString()}")
        }

        printReceivedData(response)
    }


    fun updateAccountTransactions() {
        val client = FinTs4kBankingClientForUser(bankCode, loginName, password, SimpleBankingClientCallback { tanChallenge, callback ->
            val tan: String? = readlnOrNull() // if a TAN is required, read TAN from command line, add a UI, ...
            callback.invoke(EnterTanResult(tan)) // ...  set a break point here, get TAN e.g. from your TAN app, set tan variable in debugger view and resume debugger
        })

        // simulate account transactions we retrieved last time
        val today = Clock.System.now().toLocalDateTime(TimeZone.currentSystemDefault()).date
        val lastCallToBankServer = client.getAccountData(GetAccountDataOptions(RetrieveTransactions.AccordingToRetrieveFromAndTo, retrieveTransactionsFrom = today.minusDays(90), retrieveTransactionsTo = today.minusDays(30)))

        if (lastCallToBankServer.data != null) { // now update account transactions
            val existingTransactions = lastCallToBankServer.data!!.bookedTransactions

            val updateTransactionsResponse = client.updateAccountTransactions()

            if (updateTransactionsResponse.data != null) {
                val retrievedTransactions = updateTransactionsResponse.data!!.flatMap { it.bookedTransactions }
                val newTransactions = BankingModelService().findNewTransactions(retrievedTransactions, existingTransactions)
                // `retrievedTransactions` may contain transactions we already have locally, `newTransactions` only
                // transactions that are not in `existingTransactions`
            }
        }
    }


    private fun printReceivedData(response: Response<GetAccountDataResponse>) {
        response.data?.let { data ->
            val bank = data.bank
            println("Kunde: ${bank.customerName} ${bank.accounts.size} Konten @ ${bank.bic} ${bank.bankName}")

            println()
            println("Konten:")
            bank.accounts.sortedBy { it.type }.forEach { account ->
                println("${account.identifier} ${account.productName} ${account.balance} ${account.currency}")
            }

            println()
            println("Umsätze:")
            data.bookedTransactions.forEach { transaction ->
                println("${transaction.valueDate} ${transaction.amount} ${transaction.currency} ${transaction.otherPartyName ?: ""} - ${transaction.reference}")
            }
        }
    }
}